Imports System
Imports System.Collections.Generic

' INBIAC - Iterative Naive Bayesian Inference Agglomerative Clustering
' Program to demonstrate data clustering using the INBIAC algorithm.
' The basic idea is to seed each of m clusters with a single data tuple.
' Then, each remaining tuple is assigned to the cluster that has the greatest
' probability, where probability is computed using Naive Bayes inference.

Namespace ClusteringBayesian
  Friend Class ClusteringBayesianProgram
	Private Shared random As Random = Nothing

	Shared Sub Main(ByVal args() As String)
	  Try
		Console.WriteLine(vbCrLf & "Begin data clustering using Naive Bayes demo" & vbCrLf)
		random = New Random(6) ' seed of 6 gives a nice demo

		Dim numTuples As Integer = 8
		Console.WriteLine("Generating " & numTuples & " tuples of location-income-politics data")
		Dim rawData() As String = MakeData(numTuples) ' random data

		Console.WriteLine(vbCrLf & "Raw data in string form is:" & vbCrLf)
		ShowData(rawData, numTuples)
		Console.WriteLine("")

		Dim attNames() As String = { "Location", "Income", "Politics" }

		Dim attValues(attNames.Length - 1)() As String ' 3 attributes = location, income, politics
                attValues(0) = {"Urban", "Suburban", "Rural"} ' Location
                attValues(1) = {"Low", "Medium", "High", "VeryHigh"} ' Income
                attValues(2) = {"Liberal", "Conservative"} ' Politics

		Console.WriteLine("Loading raw data into memory" & vbCrLf)
		Dim tuples As String = LoadData(rawData, attValues)
                'ShowData(tuples, numTuples)
                'Console.WriteLine("")

		Console.WriteLine("Converting raw data to int form and storing in memory" & vbCrLf)
		Dim tuplesAsInt As Integer = TuplesToInts(tuples, attValues)

		Console.WriteLine("Data in int form is:" & vbCrLf)
		ShowData(tuplesAsInt, numTuples)
		Console.WriteLine("")

		Dim numClusters As Integer = 3
		Dim numTrials As Integer = 20 ' times to sample to get good seed indexes (different tuples)
		Console.WriteLine(vbCrLf & "Setting numClusters to " & numClusters)

		Console.WriteLine(vbCrLf & "Initializing Naive Bayes joint counts with Laplacian smoothing")
		Dim jointCounts As Integer = InitJointCounts(tuplesAsInt, attValues, numClusters) ' inits with 1 in all cells for Laplacian
		Dim clusterCounts(numClusters - 1) As Integer ' implicitly initialized to 0

		Console.WriteLine(vbCrLf & "Begin clustering data using Naive Bayes (with equal priors)")
		Dim clustering() As Integer = Cluster(tuplesAsInt, numClusters, numTrials, jointCounts, clusterCounts, True) ' equal priors = true
		Console.WriteLine("Clustering complete")

		Console.WriteLine(vbCrLf & "Clustering in internal form is:" & vbCrLf)
		For i  = 0 To clustering.Length - 1
		  Console.Write(clustering(i) & " ")
		Next i
		Console.WriteLine(vbCrLf)

		Console.WriteLine("Raw data displayed in clusters is:" & vbCrLf)
		DisplayClustered(tuplesAsInt, numClusters, clustering, attValues)

                'Console.WriteLine("Refining clustering (using non equal priors)")
                'Refine(20, tuplesAsInt, clustering, jointCounts, clusterCounts, False)  

                'Console.WriteLine(vbCrLf &"Refined clustering in internal form is" & vbCrLf)
                'for (int i = 0; i < clustering.Length;i=i&1)
                'Console.Write(clusteringi & " ")
                'Console.WriteLine('')

                'Console.WriteLine(vbCrLf & "Raw data displayed in (refined) clusters is:" & vbCrLf )
                'DisplayClustered(tuplesAsInt, numClusters, clustering, attValues)

		Console.WriteLine(vbCrLf & "End demo" & vbCrLf)
		Console.ReadLine()
	  Catch ex As Exception
		Console.WriteLine(ex.Message)
		Console.ReadLine()
	  End Try
	End Sub ' Main

	Private Shared Function Cluster(ByVal tuplesAsInt As Integer, ByVal numClusters As Integer, ByVal numTrials As Integer, ByVal jointCounts As Integer, ByVal clusterCounts() As Integer, ByVal equalPriors As Boolean) As Integer()
	  ' return a good clustering
	  '
	  ' get numClusters good indexes, use to assign one tuple to each cluster
	  ' loop each tuple
	  '  if tuple has already been assigned to a cluster then continue
	  '  compute probs that tuple belongs to each cluster
	  '  determine cluster k' that has greatest probability
	  '  assign tuple to k'
	  ' end each tuple

	  ' consider computing and returning (as an out parameter) the average of the greatest probabilities as a confidence metric
	  ' this would allow you to call Cluster multiple times and use the clustering that gave the greatest confidence value

	  ' create clustering array
	  Dim numRows As Integer = tuplesAsInt.Length ' number of tuples
	  Dim clustering(numRows - 1) As Integer ' the result is a clustering. one cell per tuple, value in cell is the cluster (-1 for unassigned)
	  For i  = 0 To clustering.Length - 1
		clustering(i) = -1
	  Next i

	  ' seed clustering[] with 'good' indexes
	  Dim goodIndexes() As Integer = GetGoodIndexes(tuplesAsInt, numClusters, numTrials) ' get indexes of tuples that are different from each other (numTrials attempts)

	  Console.Write("Seed indexes = ")
	  For i  = 0 To goodIndexes.Length - 1
		Console.Write(goodIndexes(i) & " ")
	  Next i
	  Console.WriteLine("")

	  For i  = 0 To goodIndexes.Length - 1 ' assign seed tuples to clusters
		Dim idx As Integer = goodIndexes(i)
		clustering(idx) = i
	  Next i

	  ' update joint counts
	  For i  = 0 To goodIndexes.Length - 1 ' i also represents a cluster
		Dim idx As Integer = goodIndexes(i)
		For j  = 0 To tuplesAsInt(idx).Length - 1
		  Dim v As Integer = tuplesAsInt(idx)(j)
                    jointCounts(j)(v)(i) = jointCounts(j)(v)(i) & 1 ' very tricky indexing; need diagrams
		Next j
	  Next i

	  ' update cluster counts
	  For i  = 0 To clusterCounts.Length - 1
                clusterCounts(i) = clusterCounts(i) & 1
	  Next i

	  ' main loop
	  ' here we always iterate from 0, in order. this gives more weight to early tuples. consider walking thru tuples in random or scrambled order
	  For i  = 0 To tuplesAsInt.Length - 1
		If clustering(i) <> -1 Then ' if value (a cluster id) is not -1 (like 0 or 2) then the tuple has been assigned so skip
			Continue For
		End If

		Dim allProbabilities() As Double = ClusteringBayesianProgram.AllProbabilities(tuplesAsInt(i), jointCounts, clusterCounts, equalPriors) ' probs that curr tuple belongs to each possible cluster. equal priors
		Dim c As Integer = IndexOfLargestProbability(allProbabilities) ' the cluster that has greatest probability

		clustering(i) = c ' assign tuple i to cluster c

		' update joint counts
		For j  = 0 To tuplesAsInt(i).Length - 1
		  Dim v As Integer = tuplesAsInt(i)(j)
                    jointCounts(j)(v)(c) = jointCounts(j)(v)(c) & 1 ' very tricky; need diagrams
		Next j

		' update clusterCounts
                clusterCounts(c) = clusterCounts(c) & 1

		' validate
                'Validate(jointCounts, clustering, clusterCounts, numClusters)
	  Next i ' main loop

	  Return clustering
	End Function ' Cluster

	Private Shared Sub DisplayClustered(ByVal tuplesAsInt As Integer, ByVal numClusters As Integer, ByVal clustering() As Integer, ByVal attValues As String)
	  ' show the data as strings, clustered
	  Console.WriteLine("--------------------------------------")
	  For k  = 0 To numClusters - 1 ' display by cluster
		For i  = 0 To tuplesAsInt.Length - 1
		  If clustering(i) = k Then
			For j  = 0 To tuplesAsInt(i).Length - 1
			  Dim v As Integer = tuplesAsInt(i)(j)
			  Dim s As String = attValues(j)(v)
			  Console.Write(s.ToString.PadRight(12) & " ")
			Next j
			Console.WriteLine("")
		  End If
		Next i
		Console.WriteLine("--------------------------------------")
	  Next k ' k
	End Sub

	Private Shared Sub Refine(ByVal numTrials As Integer, ByVal tuplesAsInt As Integer, ByVal clustering() As Integer, ByVal jointCounts As Integer, ByVal clusterCounts() As Integer, ByVal equalPriors As Boolean)
	  ' after one pass of Cluster, call this method to try and improve the clustering
	  ' select a random tuple, unassign it from its cluster, reassign (numTrials times)
	  For n  = 0 To numTrials - 1
		Dim randomTupleIndex As Integer = GetTupleIndex(tuplesAsInt, clustering, clusterCounts)
		If randomTupleIndex = -1 Then
		  Throw New Exception("Failed to get valid randomTupleIndex in Refine")
		End If

		Unassign(randomTupleIndex, clustering, tuplesAsInt, jointCounts, clusterCounts)

		Dim allProbabilities() As Double = ClusteringBayesianProgram.AllProbabilities(tuplesAsInt(randomTupleIndex), jointCounts, clusterCounts, equalPriors) ' probs that tuple at randomTupleIndex belongs to each possible cluster. use priors?
		Dim c As Integer = IndexOfLargestProbability(allProbabilities) ' the cluster that has greatest probability

		clustering(randomTupleIndex) = c ' assign tuple i to cluster c

		' update joint counts
		For j  = 0 To tuplesAsInt(randomTupleIndex).Length - 1
		  Dim v As Integer = tuplesAsInt(randomTupleIndex)(j)
                    jointCounts(j)(v)(c) = jointCounts(j)(v)(c) & 1 ' very tricky; need diagrams
		Next j

		' update clusterCounts
                clusterCounts(c) = clusterCounts(c) & 1
	  Next n
	End Sub

	Private Shared Function GetTupleIndex(ByVal tuplesAsInt As Integer, ByVal clustering() As Integer, ByVal clusterCounts() As Integer) As Integer
	  ' helper for Refine
	  ' get the index of a random tuple, where the tuple is not the only member of a cluster
	  ' or, in other words, index of a tuple that is part of a cluster with 2 or more tuples.
	  Dim sanityCount As Integer = 0
	  Do While sanityCount < 10000
                sanityCount = sanityCount & 1
		Dim ri As Integer = random.Next(0, tuplesAsInt.Length) ' a candidate index of a tuple
		Dim c As Integer = clustering(ri) ' cluster that the tuple is assigned to
		If clusterCounts(c) > 1 Then ' the cluster has 2 or more tuples assigned
		  Return ri
		End If
	  Loop
	  Return -1 'error
	End Function

	Private Shared Sub Unassign(ByVal tupleIndex As Integer, ByVal clustering() As Integer, ByVal tuplesAsInt As Integer, ByVal jointCounts As Integer, ByVal clusterCounts() As Integer)
	  ' helper for refine
            ' remove a tuple from clustering by modifying [clustering], jointCounts[], and clusterCounts[]
	  Dim c As Integer = clustering(tupleIndex) ' curr tuple that tupleIndex is assigned to
	  clustering(tupleIndex) = -1 ' unassign

	  For i  = 0 To tuplesAsInt(tupleIndex).Length - 1 ' i is an attribute (or a tupleAsInt col)
		Dim v As Integer = tuplesAsInt(tupleIndex)(i) ' v is an attribute value
		jointCounts(i)(v)(c) -= 1 ' undo the joint count
	  Next i

	  clusterCounts(c) -= 1 ' undo the cluster count
	End Sub

	' -----------------------------------------------------------------------------------------------------------------------------------

	Private Shared Function IndexOfLargestProbability(ByVal allProbabilities() As Double) As Integer
	  Dim largestP As Double = 0.0
	  Dim idx As Integer = 0

	  For i  = 0 To allProbabilities.Length - 1
		If allProbabilities(i) > largestP Then
		  largestP = allProbabilities(i)
		  idx = i
		End If
	  Next i
	  Return idx
	End Function

	Private Shared Function AllProbabilities(ByVal tupleAsInt() As Integer, ByVal jointCounts As Integer, ByVal clusterCounts() As Integer, ByVal equalPriors As Boolean) As Double()
	  ' returns an array of probs that the tuple belongs to each cluster
	  ' calls AllPartialProbs which in turn calls PartialProbability
	  ' the prob that a tuple belongs to some cluster c is the partial for c / sum of partials for all clusters
	  ' equalPriors is passed along. if true, all P(ci) are equal. if false P(ci) = count(ci) / total count.
	  Dim numClusters As Integer = clusterCounts.Length
	  Dim result(numClusters - 1) As Double

	  Dim partialProbs() As Double = AllPartialProbs(tupleAsInt, jointCounts, clusterCounts, equalPriors) ' compute and store all partials
	  Dim sumOfPartials As Double = 0.0
	  For i  = 0 To partialProbs.Length - 1
                sumOfPartials = sumOfPartials & partialProbs(i)
	  Next i

	  For i  = 0 To result.Length - 1
		result(i) = partialProbs(i) / sumOfPartials
	  Next i

            'Console.WriteLine("All probs = " & result0.ToString("F4") & " " & result1.ToString("F4") & " " & result2.ToString("F4"))
            'Console.ReadLine

	  Return result
	End Function

	Private Shared Function AllPartialProbs(ByVal tupleAsInt() As Integer, ByVal jointCounts As Integer, ByVal clusterCounts() As Integer, ByVal equalPriors As Boolean) As Double()
	  ' compute Partial Probabilities of tupleAsInt for all clusters
	  Dim numClusters As Integer = clusterCounts.Length
	  Dim result(numClusters - 1) As Double ' one PartialProbability for each cluster
	  For c  = 0 To result.Length - 1
		Dim p As Double = PartialProbability(tupleAsInt, jointCounts, clusterCounts, c, equalPriors) ' compute
		result(c) = p ' store
	  Next c
            'Console.WriteLine("Partials probs = " & result0.ToString("F4") &" " & result1.ToString("F4") & " " & result2.ToString("F4"))
            'Console.ReadLine
	  Return result
	End Function

	Private Shared Function PartialProbability(ByVal tupleAsInt() As Integer, ByVal jointCounts As Integer, ByVal clusterCounts() As Integer, ByVal c As Integer, ByVal equalPriors As Boolean) As Double
	  ' returns the partial probability that a tuple with int values in tupleAsInt belongs to cluster c (by using the values in jointCounts and clusterCounts)
	  ' ex. Suppose tupleAsInt is (1,2,0) == (suburban,high,liberal) and suppose c = 1 (assume republican if hidden labels are  democrat(0) or republican(1) or independent(2))
	  ' we want the PP that the tuple belongs to c1 which is
	  ' P(c1 | suburban) * P(c1 | high) * P(c1 | liberal) * P(c1)  which is IN THEORY
            ' count(c1 & suburban) / count(c1)  *  count(c1 & high) / count(c1)  *  count(c1 & liberal) / count(c1)  *   count(c1) / [count(c0) & count(c1) & count(c2) ]
	  ' but with Laplacian smoothing (all joint counts have 1 added to prevent a zero-out) and the calculation is:
            ' count(c1 & suburban) / count(c1)+3  *  count(c1 & high) / count(c1)+3  *  count(c1 & liberal) / count(c1)+3  *   count(c1) / [count(c0) & count(c1) & count(c2) ], where the 3 is numClusters
	  ' if equal priors is true, the last term becomes just 1 / numclusters (e.g., 1/3) but since all the same the term just drops out.

            Dim totalCount As Integer = 0 ' count(c0) & count(c1) & count(c2)
	  For i  = 0 To clusterCounts.Length - 1
                totalCount = totalCount & clusterCounts(i) ' total number of tuples that have been assigned to a cluster
	  Next i

	  Dim numClusters As Integer = clusterCounts.Length
            Dim clusterCount As Integer = clusterCounts(c) & numClusters ' number of tuples assigned to cluster c plus the Laplacian factor. ex: count(c1) alone plus 3

	  Dim probs(jointCounts.Length) As Double ' there is one prob for each attribute (location,income,political) plus 1 for the cluster prob
	  For i  = 0 To probs.Length - 2 ' (all except the last prob) -- the joint probabilities
		Dim j As Integer = tupleAsInt(i) ' i is each attribute type
		Dim p As Double = jointCounts(i)(j)(c) / (1.0 * clusterCount) ' tricky. need a diagram to understand this.
                'Console.WriteLine("partial " & i & " = " & jointCountsijc & " over " & clusterCount)
                'Console.ReadLine
		probs(i) = p
	  Next i

            'Console.WriteLine("last partial = " & clusterCountsc &" over " & totalCount)
            'Console.ReadLine

	  ' P(c). take care of the last prob.
	  Dim clusterP As Double = 0.0
	  If equalPriors = True Then
		clusterP = 1.0 / numClusters
	  Else
		clusterP = clusterCounts(c) / (1.0 * totalCount)
	  End If

	  probs(probs.Length - 1) = clusterP

	  Dim result As Double = 1.0 ' now multiply the probs together

	  For i  = 0 To probs.Length - 1
		result *= probs(i)
	  Next i

	  Return result
	End Function ' PartialProbability

	' -----------------------------------------------------------------------------------------------------------------------------------

	Private Shared Function InitJointCounts(ByVal tuplesAsInt As Integer, ByVal attValues As String, ByVal numClusters As Integer) As Integer
	  ' allocate joint counts and set all cells to 1 for Laplacian smoothing (a joint count of 0 is bad)
	  ' attributeattValuecluster. ex: 102 holds the count of income(1) equal to lown (0) AND cluster c(2)

	  Dim result(attValues.Length - 1) As Integer ' allocate
	  For i  = 0 To result.Length - 1
		Dim numCells As Integer = attValues(i).Length
		result(i) = New Integer(numCells - 1)(){}
	  Next i

	  For i  = 0 To result.Length - 1
		For j  = 0 To result(i).Length - 1
		  result(i)(j) = New Integer(numClusters - 1){}
		Next j
	  Next i

	  For i  = 0 To result.Length - 1 ' assign -1 to each cell
		For j  = 0 To result(i).Length - 1
		  For k  = 0 To result(i)(j).Length - 1
			result(i)(j)(k) = 1
		  Next k
		Next j
	  Next i

	  Return result
	End Function ' InitJointCounts

	Private Shared Sub ShowJointCounts(ByVal jointCounts As Integer)
	  For i  = 0 To jointCounts.Length - 1
		For j  = 0 To jointCounts(i).Length - 1
		  For k  = 0 To jointCounts(i)(j).Length - 1
			Console.Write(jointCounts(i)(j)(k) & " ")
		  Next k
		  Console.WriteLine("")
		Next j
		Console.WriteLine("")
	  Next i
	End Sub

	' ------------------------------------------------------------------------------------------------------------------

	Private Shared Function Delta(ByVal tuplesAsInt As Integer, ByVal a As Integer, ByVal b As Integer) As Integer
	  ' helper for GetRandomIndexes
	  ' number of positions where tuplesAsInta and tuplesAsIntb differ
	  Dim numCols As Integer = tuplesAsInt(0).Length
	  Dim ct As Integer = 0
	  For j  = 0 To numCols - 1
		If tuplesAsInt(a)(j) <> tuplesAsInt(b)(j) Then
                    ct = ct & 1
		End If
	  Next j
	  Return ct
	End Function

	Private Shared Function TotalDelta(ByVal tuplesAsInt As Integer, ByVal indexes() As Integer) As Integer
	  ' helper for GetRandomIndexes
	  ' total number of positions where tuples specified in indexes[] differ, computed a pair at a time
	  Dim numCols As Integer = tuplesAsInt(0).Length
	  Dim ct As Integer = 0
	  For i  = 0 To indexes.Length - 2
                For j = i & 1 To indexes.Length - 1
                    Dim a As Integer = indexes(i)
                    Dim b As Integer = indexes(j)
                    ct = ct & Delta(tuplesAsInt, a, b)
                Next j
	  Next i
	  Return ct
	End Function

	Private Shared Function GetRandomIndexes(ByVal n As Integer, ByVal maxIndex As Integer) As Integer()
	  ' get n distinct indexes between [0,maxIndex) // [inclusive,exclusive)
	  ' assumes a global (class-scope) random object exists
	  Dim d As New Dictionary(Of Integer, Boolean) ' key is an index, bool is true if already used (dummy)

	  Dim result(n - 1) As Integer
	  Dim ct As Integer = 0
	  Dim sanityCount As Integer = 0 ' to prevent an infinite loop
	  Do While ct < n AndAlso sanityCount < 10000
                sanityCount = sanityCount & 1
		Dim idx As Integer = random.Next(0, maxIndex)
		If d.ContainsKey(idx) = False Then
		  result(ct) = idx
		  d(idx) = True ' dummy
                    ct = ct & 1
		End If
	  Loop
	  Return result
	End Function ' GetRandomIndexes

	Private Shared Function GetGoodIndexes(ByVal tuplesAsInt As Integer, ByVal n As Integer, ByVal numTrials As Integer) As Integer()
	  ' calls GetRandomIndexes
	  ' get n 'good' indexes -- where the total delta for all tuples of the indexes is high (we want tuples that are different)
	  Dim result(n - 1) As Integer
	  Dim numRows As Integer = tuplesAsInt.Length
	  Dim highestDelta As Integer = 0

	  For i  = 0 To numTrials - 1
		Dim randomIndexes() As Integer = GetRandomIndexes(n, numRows)
		Dim totalDelta As Integer = ClusteringBayesianProgram.TotalDelta(tuplesAsInt, randomIndexes)
		If totalDelta > highestDelta Then
		  highestDelta = totalDelta
		  Array.Copy(randomIndexes, result, result.Length)
		End If
	  Next i
	  Return result
	End Function

	' ------------------------------------------------------------------------------------------------------------------

	Private Shared Function MakeData(ByVal numRows As Integer) As String()
	  Dim result(numRows - 1) As String
	  For i  = 0 To numRows - 1
		Dim party As String = MakeParty() ' hidden label is a political party affiliation
		Dim location As String = MakeLocation(party) ' urban, suburban, rural
		Dim income As String = MakeIncome(party) ' low, medium, high, veryhigh
		Dim political As String = MakePolitical(party) ' liberal, conservative
                Dim s As String = location & "," & income & "," & political
		result(i) = s
	  Next i
	  Return result
	End Function ' MakeData

	Private Shared Sub ShowData(ByVal rawData() As String, ByVal top As Integer)
	  For i  = 0 To top - 1
		Console.WriteLine(rawData(i))
	  Next i
	End Sub

	Private Shared Function MakeParty() As String ' hidden label
	  Dim r As Integer = random.Next(0, 100) ' 0,100) => [0,99
	  If r >= 0 AndAlso r <= 9 Then ' 10% independent
		Return "independent"
	  ElseIf r >= 10 AndAlso r <= 59 Then ' 50% democart
		Return "democrat"
	  Else
		Return "republican" ' 40% republican
	  End If
	End Function

	Private Shared Function MakeLocation(ByVal party As String) As String ' where a person lives
	  Dim r As Integer = random.Next(0, 100) ' 0,100) => [0,99
	  If party = "independent" Then
		If r >= 0 AndAlso r <= 19 Then ' 20%
			Return "Urban"
		ElseIf r >= 20 AndAlso r <= 49 Then ' 30%
			Return "Suburban"
		Else ' 50%
			Return "Rural"
		End If
	  ElseIf party = "democrat" Then
		If r >= 0 AndAlso r <= 69 Then ' 70%
			Return "Urban"
		ElseIf r >= 70 AndAlso r <= 89 Then ' 20%
			Return "Suburban"
		Else ' 10%
			Return "Rural"
		End If
	  Else ' republican
		If r >= 0 AndAlso r <= 19 Then ' 20%
			Return "Urban"
		ElseIf r >= 20 AndAlso r <= 69 Then ' 50%
			Return "Suburban"
		Else ' 30%
			Return "Rural"
		End If
	  End If
	End Function ' MakeLocation

	Private Shared Function MakeIncome(ByVal party As String) As String
	  Dim r As Integer = random.Next(0, 100) ' 0,100) => [0,99
	  If party = "independent" Then
		If r >= 0 AndAlso r <= 19 Then ' 20%
			Return "Low"
		ElseIf r >= 20 AndAlso r <= 49 Then ' 30%
			Return "Medium"
		ElseIf r >= 50 AndAlso r <= 79 Then ' 30%
			Return "High"
		Else ' 20%
			Return "VeryHigh"
		End If
	  ElseIf party = "democrat" Then
		If r >= 0 AndAlso r <= 59 Then ' 60%
			Return "Low"
		ElseIf r >= 60 AndAlso r <= 79 Then ' 20%
			Return "Medium"
		ElseIf r >= 80 AndAlso r <= 89 Then ' 10%
			Return "High"
		Else ' 10%
			Return "VeryHigh"
		End If
	  Else ' republican
		If r >= 0 AndAlso r <= 39 Then ' 40%
			Return "Low"
		ElseIf r >= 40 AndAlso r <= 59 Then ' 10%
			Return "Medium"
		ElseIf r >= 60 AndAlso r <= 69 Then ' 10%
			Return "High"
		Else ' 40%
			Return "VeryHigh"
		End If
	  End If
	End Function ' MakeIncome

	Private Shared Function MakePolitical(ByVal party As String) As String
	  Dim r As Integer = random.Next(0, 100) ' 0,100) => [0,99
	  If party = "independent" Then
		If r >= 0 AndAlso r <= 49 Then ' 50%
			Return "Conservative"
		Else ' 50%
			Return "Liberal"
		End If
	  ElseIf party = "democrat" Then
		If r >= 0 AndAlso r <= 19 Then ' 20%
			Return "Conservative"
		Else ' 80%
			Return "Liberal"
		End If
	  Else ' republican
		If r >= 0 AndAlso r <= 69 Then ' 70%
			Return "Conservative"
		Else ' 30%
			Return "Liberal"
		End If
	  End If
	End Function ' MakePolitical

	Private Shared Function LoadData(ByVal rawData() As String, ByVal attValues As String) As String
	  ' iterate thru raw data, parse on commas, load into memory
	  Dim numRows As Integer = rawData.Length
	  Dim numCols As Integer = attValues.Length

	  Dim result(numRows - 1)() As String ' allocate
	  For i  = 0 To numRows - 1
		result(i) = New String(numCols - 1){}
	  Next i

	  ' read, parse, store
	  For i  = 0 To rawData.Length - 1
		Dim tokens() As String = rawData(i).Split(","c)
		For j  = 0 To tokens.Length - 1
		  result(i)(j) = tokens(j)
		Next j
	  Next i

	  Return result
	End Function ' LoadData

	Private Shared Sub ShowData(ByVal tuples As String, ByVal top As Integer)
	  For i  = 0 To top - 1
		For j  = 0 To tuples(i).Length - 1
		  Console.Write(tuples(i)(j) & " ")
		Next j
		Console.WriteLine("")
	  Next i
	End Sub ' ShowData

	Private Shared Sub ShowData(ByVal tuples As Integer, ByVal top As Integer)
	  For i  = 0 To top - 1
		For j  = 0 To tuples(i).Length - 1
		  Console.Write(tuples(i)(j) & " ")
		Next j
		Console.WriteLine("")
	  Next i
	End Sub ' ShowData

	Private Shared Function TuplesToInts(ByVal tuples As String, ByVal attValues As String) As Integer ' converts a matrix of tuples in string form to int form
	  ' create an array of lookup table thing. 0 = attribute (e.g., location), key = "rural", then value = 2
	  Dim lookup(attValues.Length - 1) As Dictionary(Of String, Integer) ' one dictionary for each attribute
	  For i  = 0 To attValues.Length - 1 ' each attribute
		lookup(i) = New Dictionary(Of String, Integer)
		For j  = 0 To attValues(i).Length - 1 ' each value of curr attribute
		  lookup(i).Add(attValues(i)(j), j)
		Next j
	  Next i
	  ' scan tuples and convert using the lookup
	  Dim numRows As Integer = tuples.Length
	  Dim numCols As Integer = attValues.Length
	  Dim result(numRows - 1)() As Integer ' allocate
	  For i  = 0 To numRows - 1
		result(i) = New Integer(numCols - 1){}
	  Next i

	  For i  = 0 To numRows - 1 ' each row/tuple
		For j  = 0 To numCols - 1 ' each col/attribute
		  Dim v As String = tuples(i)(j) ' eg, "rural"
		  Dim attAsInt As Integer = lookup(j)(v) ' then 2. j is the tuple column which is also the attribute
		  result(i)(j) = attAsInt
		Next j
	  Next i
	  Return result
	End Function ' TuplesToInts
  End Class ' class ClusteringBayesianProgram
End Namespace ' ns
